import logging
from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Any, Optional

from bson import ObjectId
from pymongo import ASCENDING, DESCENDING
from pymongo.collection import Collection

from jarvis.config import get_config
from jarvis.db import DBManager

logger = logging.getLogger("mongodb")

sort_lookup = {
    "asc": ASCENDING,
    "ascending": ASCENDING,
    "desc": DESCENDING,
    "descending": DESCENDING,
}

coll_lookup = {
    "Autoreact": "autoreact",
    "Ban": "bans",
    "Config": "config",
    "Lock": "locks",
    "Kick": "kicks",
    "Mute": "mutes",
    "Purge": "purges",
    "Setting": "settings",
    "Starboard": "starboard",
    "Star": "stars",
    "Unban": "unbans",
    "Warning": "warns",
}

db_instance = DBManager(get_config().mongo).mongo.jarvis


#################
#  Core Classes #
#################


@dataclass
class MongoSort:
    direction: int
    key: Any

    def __init__(self, direction, key):
        if type(direction) is str:
            direction = sort_lookup[direction]
        self.direction = direction
        self.key = key

    def as_tuple(self):
        return (self.key, self.direction)


@dataclass
class MongoObject:
    """
    A MongoDB object

    :param _id: MongoDB ObjectId, not initialized by default
    """

    _id: ObjectId = field(default=None, init=False)

    def __post_init__(self):
        self.coll = db_instance[coll_lookup[type(self).__name__]]

    def to_dict(self):
        return asdict(self)

    def insert(self) -> ObjectId:
        """
        Inserts the object into the database

        :param collection: Collection to insert object into
        :return: Inserted object ObjectId
        """
        to_insert = self.to_dict()
        if not to_insert["_id"]:
            _ = to_insert.pop("_id")
        try:
            result = self.coll.insert_one(to_insert)
        except Exception as e:
            logger.error(f"Failed to insert {type(self).__name__}", e)
        else:
            id = result.inserted_id
            self._id = id
            return id

    def delete(self) -> bool:
        """
        Deletes an object from the database

        :param collection: Collection to delete object from
        :return: If delete was successful
        """
        to_delete = self.to_dict()
        search = {"_id": to_delete["_id"]}
        try:
            result = self.coll.delete_one(search)
        except Exception as e:
            logger.error(f"Failed to delete {type(self).__name__}", e)
        else:
            return result.deleted_count > 0

    @classmethod
    def delete_many(cls, **kwargs) -> bool:
        try:
            coll = db_instance[coll_lookup[cls.__name__]]
            result = coll.delete_many(kwargs)
        except Exception as e:
            logger.error(f"Failed to delete {cls.__name__}s", e)
        else:
            return result.deleted_count > 0

    def update(self) -> bool:
        """
        Updates an object in the database

        :param collection: Collection to update
        :return: If update was successful
        """
        to_update = self.to_dict()
        search = {"_id": to_update["_id"]}
        try:
            result = self.coll.update_one(
                search, {"$set": to_update}, upsert=True
            )
        except Exception as e:
            logger.error(f"Failed to update {type(self).__name__}", e)
        else:
            return result.modified_count > 0

    @classmethod
    def get(cls, **kwargs) -> Optional[object]:
        """
        Get an object from MongoDB

        :param collection: Collection to query
        :return: Optional object
        """
        try:
            sort = None
            if "sort" in kwargs:
                sort = kwargs.pop("sort")
                if type(sort) is list:
                    sort = [x.as_tuple() for x in sort]
                else:
                    sort = [sort.as_tuple()]
            coll = db_instance[coll_lookup[cls.__name__]]
            result = coll.find_one(kwargs, sort=sort)
        except Exception as e:
            logger.error(f"Failed to get {cls.__name__}", e)
        else:
            if result:
                _id = result.pop("_id")
                r = cls(**result)
                r._id = _id
                return r
            return None

    @classmethod
    def get_many(cls, **kwargs) -> Optional[list]:
        """
        Gets objects from MongoDB

        :param collection: Collection to query
        :return: Optional object
        """
        global db_instance
        try:
            sort = None
            if "sort" in kwargs:
                sort = kwargs.pop("sort")
                if type(sort) is list:
                    sort = [x.as_tuple() for x in sort]
                else:
                    sort = [sort.as_tuple()]
            coll = db_instance[coll_lookup[cls.__name__]]
            result = coll.find(kwargs, sort=sort)
        except Exception as e:
            logger.error(f"Failed to get {cls.__name__}", e)
        else:
            if result:
                r_list = []
                for r_item in result:
                    _id = r_item.pop("_id")
                    r = cls(**r_item)
                    r._id = _id
                    r_list.append(r)
                return r_list
            return []


@dataclass
class ActiveObject:
    """
    A type of Mongo object that can be active

    :param active: If object is active
    """

    active: bool

    @classmethod
    def get_active(cls, **kwargs) -> list:
        """
        Gets a list of active objects

        :param collection: Collection to query
        :return: List of active objects
        """
        kwargs.update({"active": True})
        try:
            sort = None
            if "sort" in kwargs:
                sort = kwargs.pop("sort")
                if type(sort) is list:
                    sort = [x.as_tuple() for x in sort]
                else:
                    sort = [sort.as_tuple()]
            coll = db_instance[coll_lookup[cls.__name__]]
            result = coll.find(kwargs, sort=sort)
        except Exception as e:
            logger.error(f"Failed to get {cls.__name__}", e)
        else:
            if result:
                r_list = []
                for r_item in result:
                    _id = r_item.pop("_id")
                    r = cls(**r_item)
                    r._id = _id
                    r_list.append(r)
                return r_list
            return []


####################
# Database Objects #
####################


@dataclass
class Autopurge(MongoObject):
    """
    Channel Autoreact object

    :param _id: MongoDB ID
    :param guild: ID of guild
    :param channel: ID of channel
    :param delay: Purge delay
    :param admin: ID of admin who added autoreact
    :param created_at: Time the autoreact was created
    """

    guild: int
    channel: int
    delay: int
    admin: int
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Autoreact(MongoObject):
    """
    Channel Autoreact object

    :param _id: MongoDB ID
    :param guild: ID of guild
    :param channel: ID of channel
    :param reactions: List of reactions
    :param admin: ID of admin who added autoreact
    :param created_at: Time the autoreact was created
    """

    guild: int
    channel: int
    reactions: list
    admin: int
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Ban(MongoObject, ActiveObject):
    """
    User Ban object

    :param _id: MongoDB ID
    :param active: If the ban is active
    :param admin: ID of admin who banned the user
    :param user: ID of banned user
    :param username: Username of banned user
    :param discrim: Discriminator of banned user
    :param duration: Duration of ban
    :param guild: ID of guild that user belonged to
    :param type: Type of ban
    :param reason: Reason for the ban
    :param created_at: Time the ban happened
    """

    admin: int
    user: int
    username: str
    discrim: int
    duration: int
    guild: int
    type: str
    reason: str
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Config(MongoObject):
    """
    J.A.R.V.I.S. Config object

    :param _id: MongoDB ID
    :param key: Config key
    :param value: Config value
    """

    key: str
    value: any


@dataclass
class Joke(MongoObject):
    """
    Joke object

    :param _id: MongoDB ID
    :param id: Reddit ID
    :param body: Joke Body
    :param title: Joke Title
    :param created_utc: Created at
    :param over_18: Is NSFW
    :param score: Reddit Score
    """

    id: str
    body: str
    title: str
    created_utc: datetime
    over_18: bool
    score: int


@dataclass
class Kick(MongoObject):
    """
    User Kick object

    :param _id: MongoDB ID
    :param user: Kicked User ID
    :param reason: Kick reason
    :param admin: ID of admin who kicked user
    :param created_at: Time user was kicked
    :param guild: Guild ID
    """

    admin: int
    guild: int
    reason: str
    user: int
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Lock(MongoObject, ActiveObject):
    """
    Channel Lock object

    :param _id: MongoDB ID
    :param active: If the lock is active
    :param admin: ID of admin who locked channel
    :param channel: ID of locked channel
    :param duration: Duration of lock
    :param guild: ID of guild that channel belongs to
    :param reason: Reason for the lock
    :param created_at: Time the lock happened
    """

    admin: int
    channel: int
    duration: int
    guild: int
    reason: str
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Mute(MongoObject, ActiveObject):
    """
    User Mute object

    :param _id: MongoDB ID
    :param active: If the mute is active
    :param admin: ID of admin who muted the user
    :param user: ID of muted user
    :param duration: Duration of mute
    :param guild: ID of guild that user belongs to
    :param reason: Reason for the mute
    :param created_at: Time the mute happened
    """

    admin: int
    user: int
    duration: int
    guild: int
    reason: str
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Purge(MongoObject):
    """
    Channel Purge object

    :param _id: MongoDB ID
    :param admin: ID of admin who purged messages
    :param channel: ID of purged channel
    :param guild: ID of guild that channel belongs to
    :param count: Number of purged messages
    :param created_at: Time the purge happened
    """

    admin: int
    channel: int
    guild: int
    count: int
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Setting(MongoObject):
    """
    Guild Setting object

    :param _id: MongoDB ID
    :param guild: ID of guild
    :param setting: Setting key
    :param value: Setting value
    """

    guild: int
    setting: str
    value: any


@dataclass
class Star(MongoObject):
    """
    Starboard Star object

    :param _id: MongoDB ID
    :param index: Starboard star index
    :param message: Star Message ID
    :param channel: Star Channel ID
    :param starboard: Starboard Channel ID
    :param guild: Starboard Guild ID
    :param admin: ID of admin who created star
    :param star: ID of the star message
    :param created_at: Time created
    """

    index: int
    message: int
    channel: int
    starboard: int
    guild: int
    admin: int
    star: int
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Starboard(MongoObject):
    """
    Channel Starboard object

    :param _id: MongoDB ID
    :param channel: Starboard Channel ID
    :param guild: Starboard Guild ID
    :param admin: ID of admin who created starboard
    :param created_at: Time created
    """

    channel: int
    guild: int
    admin: int
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Unban(MongoObject):
    """
    Guild Unban object

    :param _id: MongoDB ID
    :param user: User ID
    :param username: User Username
    :param discrim: User Discriminator
    :param guild: Guild ID
    :param admin: Admin who unbanned user
    :param reason: Reason for unban
    :param created_at: Time created
    """

    user: int
    username: str
    discrim: int
    guild: int
    admin: int
    reason: str
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Warning(MongoObject, ActiveObject):
    """
    User Warning object

    :param _id: MongoDB ID
    :param active: If the warning is active
    :param admin: ID of admin who warned the user
    :param created_at: Time the warning happened
    :param duration: Duration of warning
    :param guild: ID of guild that user belongs to
    :param reason: Reason for the warning
    :param user: ID of warned user
    """

    admin: int
    user: int
    duration: int
    guild: int
    reason: str
    created_at: datetime = field(default_factory=datetime.utcnow)
